import json
import logging
from collections import defaultdict
from datetime import datetime
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

import pytz
from peek_core_docdb._private.storage.DocDbCompilerQueue import \
    DocDbCompilerQueue
from peek_core_docdb._private.storage.DocDbDocument import DocDbDocument
from peek_core_docdb._private.storage.DocDbDocumentTypeTuple import (
    DocDbDocumentTypeTuple,
)
from peek_core_docdb._private.storage.DocDbModelSet import DocDbModelSet
from peek_core_docdb._private.storage.DocDbPropertyTuple import \
    DocDbPropertyTuple
from peek_core_docdb._private.worker.tasks._CalcChunkKey import makeChunkKey
from peek_core_docdb.tuples.ImportDocumentTuple import ImportDocumentTuple
from sqlalchemy import and_
from sqlalchemy import bindparam
from sqlalchemy import select
from txcelery.defer import DeferrableTask

from peek_plugin_base.worker import CeleryDbConn
from peek_plugin_base.worker.CeleryApp import celeryApp
from vortex.Jsonable import Jsonable
from vortex.Payload import Payload


logger = logging.getLogger(__name__)


# We need to insert the into the following tables:
# DocDbDocument - or update it's details if required
# DocDbIndex - The index of the keywords for the object
# DocDbDocumentRoute - delete old importGroupHash
# DocDbDocumentRoute - insert the new routes


@DeferrableTask
@celeryApp.task(bind=True)
def removeDocumentTask(self, modelSetKey: str, keys: List[str]) -> None:
    pass


@DeferrableTask
@celeryApp.task(bind=True)
def createOrUpdateDocuments(self, documentsEncodedPayload: bytes) -> None:
    startTime = datetime.now(pytz.utc)
    # Decode arguments
    newDocuments: List[ImportDocumentTuple] = (
        Payload().fromEncodedPayload(documentsEncodedPayload).tuples
    )

    _validateNewDocuments(newDocuments)

    modelSetIdByKey = _loadModelSets()

    # Do the import
    try:

        documentByModelKey = defaultdict(list)
        for doc in newDocuments:
            documentByModelKey[doc.modelSetKey].append(doc)

        for modelSetKey, docs in documentByModelKey.items():
            modelSetId = modelSetIdByKey.get(modelSetKey)
            if modelSetId is None:
                modelSetId = _makeModelSet(modelSetKey)
                modelSetIdByKey[modelSetKey] = modelSetId

            docTypeIdsByName = _prepareLookups(docs, modelSetId)
            _insertOrUpdateObjects(docs, modelSetId, docTypeIdsByName)

        logger.info(
            "Imported %s Documents in %s",
            len(newDocuments),
            datetime.now(pytz.utc) - startTime,
        )

    except Exception as e:
        logger.debug("Retrying import docDb objects, %s", e)
        raise self.retry(exc=e, countdown=3)


def _validateNewDocuments(newDocuments: List[ImportDocumentTuple]) -> None:
    for doc in newDocuments:
        if not doc.key:
            raise Exception("key is empty for %s" % doc)

        if not doc.modelSetKey:
            raise Exception("modelSetKey is empty for %s" % doc)

        if not doc.documentTypeKey:
            raise Exception("documentTypeKey is empty for %s" % doc)

        # if not doc.document:
        #     raise Exception("document is empty for %s" % doc)


def _loadModelSets() -> Dict[str, int]:
    # Get the model set
    engine = CeleryDbConn.getDbEngine()
    conn = engine.connect()
    try:
        modelSetTable = DocDbModelSet.__table__
        results = list(
            conn.execute(
                select(columns=[modelSetTable.c.id, modelSetTable.c.key]))
        )
        modelSetIdByKey = {o.key: o.id for o in results}
        del results

    finally:
        conn.close()
    return modelSetIdByKey


def _makeModelSet(modelSetKey: str) -> int:
    # Get the model set
    dbSession = CeleryDbConn.getDbSession()
    try:
        newItem = DocDbModelSet(key=modelSetKey, name=modelSetKey)
        dbSession.add(newItem)
        dbSession.commit()
        return newItem.id

    finally:
        dbSession.close()


def _prepareLookups(
    newDocuments: List[ImportDocumentTuple], modelSetId: int
) -> Dict[str, int]:
    """Check Or Insert Search Properties

    Make sure the search properties exist.

    """

    dbSession = CeleryDbConn.getDbSession()

    startTime = datetime.now(pytz.utc)

    try:

        docTypeNames = set()
        propertyNames = set()

        for o in newDocuments:
            o.document["key"] = o.key
            o.documentTypeKey = o.documentTypeKey.lower()
            docTypeNames.add(o.documentTypeKey)

            if o.document:
                propertyNames.update([s.lower() for s in o.document])

        # Prepare Properties
        dbProps = (
            dbSession.query(DocDbPropertyTuple)
                .filter(DocDbPropertyTuple.modelSetId == modelSetId)
                .all()
        )

        propertyNames -= set([o.name for o in dbProps])

        if propertyNames:
            for newPropName in propertyNames:
                dbSession.add(
                    DocDbPropertyTuple(
                        name=newPropName, title=newPropName,
                        modelSetId=modelSetId
                    )
                )

            dbSession.commit()

        del dbProps
        del propertyNames

        # Prepare Object Types
        dbObjectTypes = (
            dbSession.query(DocDbDocumentTypeTuple)
                .filter(DocDbDocumentTypeTuple.modelSetId == modelSetId)
                .all()
        )
        docTypeNames -= set([o.name for o in dbObjectTypes])

        if not docTypeNames:
            docTypeIdsByName = {o.name: o.id for o in dbObjectTypes}

        else:
            for newType in docTypeNames:
                dbSession.add(
                    DocDbDocumentTypeTuple(
                        name=newType, title=newType, modelSetId=modelSetId
                    )
                )

            dbSession.commit()

            dbObjectTypes = dbSession.query(DocDbDocumentTypeTuple).all()
            docTypeIdsByName = {o.name: o.id for o in dbObjectTypes}

        logger.debug("Prepared lookups in %s",
            (datetime.now(pytz.utc) - startTime))

        return docTypeIdsByName

    except Exception as e:
        dbSession.rollback()
        raise

    finally:
        dbSession.close()


def _insertOrUpdateObjects(
    newDocuments: List[ImportDocumentTuple],
    modelSetId: int,
    docTypeIdsByName: Dict[str, int],
) -> None:
    """Insert or Update Objects

    1) Find objects and update them
    2) Insert object if the are missing

    """

    documentTable = DocDbDocument.__table__
    queueTable = DocDbCompilerQueue.__table__

    startTime = datetime.now(pytz.utc)

    engine = CeleryDbConn.getDbEngine()
    conn = engine.connect()
    transaction = conn.begin()

    try:
        dontDeleteObjectIds = []
        objectIdByKey: Dict[str, int] = {}

        objectKeys = [o.key for o in newDocuments]
        chunkKeysForQueue: Set[Tuple[str, str]] = set()

        # Query existing objects
        results = list(
            conn.execute(
                select(
                    columns=[
                        documentTable.c.id,
                        documentTable.c.key,
                        documentTable.c.chunkKey,
                        documentTable.c.documentJson,
                    ],
                    whereclause=and_(
                        documentTable.c.key.in_(objectKeys),
                        documentTable.c.modelSetId == modelSetId,
                    ),
                )
            )
        )

        foundObjectByKey = {o.key: o for o in results}
        del results

        # Get the IDs that we need
        newIdGen = CeleryDbConn.prefetchDeclarativeIds(
            DocDbDocument, len(newDocuments) - len(foundObjectByKey)
        )

        # Create state arrays
        inserts = []
        updates = []
        processedKeys = set()

        # Work out which objects have been updated or need inserting
        for importDocument in newDocuments:
            if importDocument.key in processedKeys:
                raise Exception(
                    "Key %s exists in import data twice" % importDocument.key
                )
            processedKeys.add(importDocument.key)

            existingObject = foundObjectByKey.get(importDocument.key)
            importDocumentTypeId = docTypeIdsByName[
                importDocument.documentTypeKey]

            packedJsonDict = {
                k: v
                for k, v in importDocument.document.items()
                if v is not None and v is not ""
            }  # 0 / false allowed
            packedJsonDict["_dtid"] = importDocumentTypeId
            packedJsonDict["_msid"] = modelSetId

            documentJson = Jsonable().toJsonField(packedJsonDict, {})
            documentJson = json.dumps(documentJson)
            # Work out if we need to update the object type
            if existingObject:
                updates.append(
                    dict(
                        b_id=existingObject.id,
                        b_typeId=importDocumentTypeId,
                        b_documentJson=documentJson,
                    )
                )
                dontDeleteObjectIds.append(existingObject.id)

            else:
                id_ = next(newIdGen)
                existingObject = DocDbDocument(
                    id=id_,
                    modelSetId=modelSetId,
                    documentTypeId=importDocumentTypeId,
                    key=importDocument.key,
                    importGroupHash=importDocument.importGroupHash,
                    chunkKey=makeChunkKey(
                        importDocument.modelSetKey, importDocument.key
                    ),
                    documentJson=documentJson,
                )
                inserts.append(existingObject.tupleToSqlaBulkInsertDict())

            objectIdByKey[existingObject.key] = existingObject.id
            chunkKeysForQueue.add((modelSetId, existingObject.chunkKey))

        # Insert the DocDb Objects
        if inserts:
            conn.execute(documentTable.insert(), inserts)

        if updates:
            stmt = (
                documentTable.update()
                    .where(documentTable.c.id == bindparam("b_id"))
                    .values(
                    documentTypeId=bindparam("b_typeId"),
                    documentJson=bindparam("b_documentJson"),
                )
            )
            conn.execute(stmt, updates)

        if chunkKeysForQueue:
            conn.execute(
                queueTable.insert(),
                [dict(modelSetId=m, chunkKey=c) for m, c in chunkKeysForQueue],
            )

        if inserts or updates or chunkKeysForQueue:
            transaction.commit()
        else:
            transaction.rollback()

        logger.debug(
            "Inserted %s updated %s queued %s chunks in %s",
            len(inserts),
            len(updates),
            len(chunkKeysForQueue),
            (datetime.now(pytz.utc) - startTime),
        )

    except Exception:
        transaction.rollback()
        raise

    finally:
        conn.close()



from peek_plugin_base.storage.DbConnection import DbSessionCreator
from peek_core_docdb._private.PluginNames import docDbFilt
from peek_core_docdb._private.PluginNames import docDbObservableName
from peek_core_docdb._private.server.tuple_providers.DocumentPropertyTupleProvider import (
    DocumentPropertyTupleProvider,
)
from peek_core_docdb._private.server.tuple_providers.DocumentTypeTupleProvider import (
    DocumentTypeTupleProvider,
)
from peek_core_docdb._private.server.tuple_providers.ModelSetTupleProvider import (
    ModelSetTupleProvider,
)
from peek_core_docdb._private.storage.DocDbDocument import DocDbDocument
from peek_core_docdb._private.storage.DocDbDocumentTypeTuple import (
    DocDbDocumentTypeTuple,
)
from peek_core_docdb._private.storage.DocDbModelSet import DocDbModelSet
from peek_core_docdb._private.storage.DocDbPropertyTuple import DocDbPropertyTuple
from peek_core_docdb._private.tuples.AdminStatusTuple import AdminStatusTuple
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler
from .controller.StatusController import StatusController
from .tuple_providers.AdminStatusTupleProvider import AdminStatusTupleProvider
from .tuple_providers.DocumentTupleProvider import DocumentTupleProvider


def makeTupleDataObservableHandler(
    dbSessionCreator: DbSessionCreator, statusController: StatusController
):
    """ " Make Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param dbSessionCreator: A function that returns a SQLAlchemy session when called
    :param statusController:

    :return: An instance of :code:`TupleDataObservableHandler`

    """
    tupleObservable = TupleDataObservableHandler(
        observableName=docDbObservableName, additionalFilt=docDbFilt
    )

    # Register TupleProviders here
    tupleObservable.addTupleProvider(
        DocDbDocument.tupleName(), DocumentTupleProvider(dbSessionCreator)
    )

    # Admin status tuple
    tupleObservable.addTupleProvider(
        AdminStatusTuple.tupleName(), AdminStatusTupleProvider(statusController)
    )

    # Document Type Tuple
    tupleObservable.addTupleProvider(
        DocDbDocumentTypeTuple.tupleName(), DocumentTypeTupleProvider(dbSessionCreator)
    )

    # Document Property Tuple
    tupleObservable.addTupleProvider(
        DocDbPropertyTuple.tupleName(), DocumentPropertyTupleProvider(dbSessionCreator)
    )

    # Model Set Tuple
    tupleObservable.addTupleProvider(
        DocDbModelSet.tupleName(), ModelSetTupleProvider(dbSessionCreator)
    )

    return tupleObservable



from abc import ABCMeta, abstractmethod
from typing import List

from twisted.internet.defer import Deferred


class DocDbApiABC(metaclass=ABCMeta):
    @abstractmethod
    def createOrUpdateDocuments(self, documentsEncodedPayload: bytes) -> Deferred:
        """Create or Update Documents

        Add new documents to the document db

        :param documentsEncodedPayload: An encoded payload containing :code:`List[DocumentTuple]`
        :return: A deferred that fires when the creates or updates are complete

        """

    @abstractmethod
    def deleteDocuments(self, modelSetKey: str, keys: List[str]) -> Deferred:
        """Delete Documents

        Delete documents from the document db.

        :param modelSetKey: the model set key to delete documents from
        :param keys: A list of keys for documents to delete
        :return: A deferred that fires when the delete is complete

        """



from vortex.handler.TupleActionProcessor import TupleActionProcessor

from peek_core_docdb._private.PluginNames import docDbFilt
from peek_core_docdb._private.PluginNames import docDbActionProcessorName
from .controller.MainController import MainController


def makeTupleActionProcessorHandler(mainController: MainController):
    processor = TupleActionProcessor(
        tupleActionProcessorName=docDbActionProcessorName,
        additionalFilt=docDbFilt,
        defaultDelegate=mainController,
    )
    return processor



docDbFilt = {"plugin": "peek_core_docdb"}
docDbTuplePrefix = "peek_core_docdb."
docDbObservableName = "peek_core_docdb"
docDbActionProcessorName = "peek_core_docdb"
docDbTupleOfflineServiceName = "peek_core_docdb"



"""Increased chunk size

Peek Plugin Database Migration Script

Revision ID: a423a783700f
Revises: c1d2d5475c64
Create Date: 2018-07-04 21:23:03.688758

"""

# revision identifiers, used by Alembic.
from sqlalchemy import MetaData
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from peek_core_docdb._private.worker.tasks._CalcChunkKey import makeChunkKey

revision = "a423a783700f"
down_revision = "c1d2d5475c64"
branch_labels = None
depends_on = None

from alembic import op

from sqlalchemy import Column
from sqlalchemy import Integer, String

import logging

logger = logging.getLogger(__name__)

__DeclarativeBase = declarative_base(metadata=MetaData(schema="core_docdb"))


class __DocDbModelSet(__DeclarativeBase):
    __tablename__ = "DocDbModelSet"

    id = Column(Integer, primary_key=True, autoincrement=True)
    key = Column(String, nullable=False, unique=True)


class __DocDbDocument(__DeclarativeBase):
    __tablename__ = "DocDbDocument"

    id = Column(Integer, primary_key=True, autoincrement=True)
    modelSetId = Column(Integer)
    key = Column(String)
    chunkKey = Column(String)


def _loadSearchObjects(session):
    FETCH_SIZE = 5000
    lastOffset = 0
    while True:
        rows = (
            session.query(__DocDbDocument)
            .order_by(__DocDbDocument.id)
            .offset(lastOffset)
            .limit(FETCH_SIZE)
            .yield_per(FETCH_SIZE)
            .all()
        )
        if not rows:
            return
        logger.info(
            "Loading %s-%s for %s",
            lastOffset,
            lastOffset + FETCH_SIZE,
            __DocDbDocument.__name__,
        )
        yield rows
        lastOffset += FETCH_SIZE


def upgrade():
    bind = op.get_bind()
    session = sessionmaker()(bind=bind)

    modelKeysById = {o.id: o.key for o in session.query(__DocDbModelSet)}

    for rows in _loadSearchObjects(session):
        for item in rows:
            item.chunkKey = makeChunkKey(modelKeysById[item.modelSetId], item.key)
        session.commit()
        session.expunge_all()

    session.close()

    op.execute('TRUNCATE TABLE core_docdb."DocDbEncodedChunkTuple" ')
    op.execute('TRUNCATE TABLE core_docdb."DocDbChunkQueue" ')

    op.execute(
        """INSERT INTO core_docdb."DocDbChunkQueue"
                            ("modelSetId", "chunkKey")
                            SELECT DISTINCT "modelSetId", "chunkKey"
                            FROM core_docdb."DocDbDocument"
                         """
    )


def downgrade():
    raise NotImplementedError("Downgrade not implemented")



import logging
from typing import Union

from twisted.internet.defer import Deferred, inlineCallbacks

from peek_core_docdb._private.server.controller.StatusController import StatusController
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)


class AdminStatusTupleProvider(TuplesProviderABC):
    def __init__(self, statusController: StatusController):
        self._statusController = statusController

    @inlineCallbacks
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:
        tuples = [self._statusController.status]

        payloadEnvelope = yield Payload(filt, tuples=tuples).makePayloadEnvelopeDefer()
        vortexMsg = yield payloadEnvelope.toVortexMsgDefer()
        return vortexMsg



from typing import Dict

from peek_core_docdb._private.PluginNames import docDbTuplePrefix
from peek_core_docdb._private.storage.DocDbDocumentTypeTuple import (
    DocDbDocumentTypeTuple,
)
from peek_core_docdb._private.storage.DocDbModelSet import DocDbModelSet
from vortex.Tuple import Tuple, addTupleType, TupleField


@addTupleType
class DocumentTuple(Tuple):
    """Document Tuple

    This tuple is the publicly exposed Document

    """

    __tupleType__ = docDbTuplePrefix + "DocumentTuple"

    #:  The unique key of this document
    key: str = TupleField()

    #:  The model set of this document
    modelSet: DocDbModelSet = TupleField()

    #:  The document type
    documentType: DocDbDocumentTypeTuple = TupleField()

    #:  The document data
    document: Dict = TupleField()



from peek_core_docdb._private.server.admin_backend.EditDocumentPropertyHandler import (
    makeDocumentPropertyHandler,
)
from peek_core_docdb._private.server.admin_backend.EditDocumentTypeHandler import (
    makeDocumentTypeHandler,
)
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler
from .SettingPropertyHandler import makeSettingPropertyHandler
from .ViewDocumentHandler import makeDocumentTableHandler


def makeAdminBackendHandlers(
    tupleObservable: TupleDataObservableHandler, dbSessionCreator
):
    yield makeDocumentTableHandler(tupleObservable, dbSessionCreator)

    yield makeSettingPropertyHandler(dbSessionCreator)

    yield makeDocumentPropertyHandler(tupleObservable, dbSessionCreator)

    yield makeDocumentTypeHandler(tupleObservable, dbSessionCreator)



import logging

from peek_abstract_chunked_index.private.client.controller.ACICacheControllerABC import (
    ACICacheControllerABC,
)
from peek_core_docdb._private.PluginNames import docDbFilt
from peek_core_docdb._private.server.client_handlers.ClientChunkLoadRpc import (
    ClientChunkLoadRpc,
)
from peek_core_docdb._private.storage.DocDbEncodedChunk import DocDbEncodedChunk
from peek_core_docdb._private.tuples.DocumentUpdateDateTuple import (
    DocumentUpdateDateTuple,
)

logger = logging.getLogger(__name__)

clientDocumentUpdateFromServerFilt = dict(key="clientDocumentUpdateFromServer")
clientDocumentUpdateFromServerFilt.update(docDbFilt)


class DocumentCacheController(ACICacheControllerABC):
    """Document Cache Controller

    The Document cache controller stores all the chunks in memory,
    allowing fast access from the mobile and desktop devices.

    """

    _ChunkedTuple = DocDbEncodedChunk
    _UpdateDateTupleABC = DocumentUpdateDateTuple
    _chunkLoadRpcMethod = ClientChunkLoadRpc.loadDocumentChunks
    _chunkIndexDeltaRpcMethod = ClientChunkLoadRpc.loadDocumentIndexDelta
    _updateFromLogicFilt = clientDocumentUpdateFromServerFilt
    _logger = logger



from sqlalchemy import Column, Index, ForeignKey
from sqlalchemy import Integer, String
from sqlalchemy.orm import relationship

from peek_core_docdb._private.PluginNames import docDbTuplePrefix
from peek_core_docdb._private.storage.DeclarativeBase import DeclarativeBase
from peek_core_docdb._private.storage.DocDbModelSet import DocDbModelSet
from vortex.Tuple import Tuple, addTupleType


@addTupleType
class DocDbDocumentTypeTuple(Tuple, DeclarativeBase):
    __tupleType__ = docDbTuplePrefix + "DocDbDocumentTypeTuple"
    __tablename__ = "DocDbDocumentType"

    id = Column(Integer, primary_key=True, autoincrement=True)

    #:  The model set for this document
    modelSetId = Column(
        Integer, ForeignKey("DocDbModelSet.id", ondelete="CASCADE"), nullable=False
    )
    modelSet = relationship(DocDbModelSet)

    name = Column(String, nullable=False)
    title = Column(String, nullable=False)

    __table_args__ = (
        Index("idx_DocType_model_name", modelSetId, name, unique=True),
        Index("idx_DocType_model_title", modelSetId, title, unique=True),
    )



from typing import Dict

from peek_core_docdb._private.PluginNames import docDbTuplePrefix
from vortex.Tuple import addTupleType, TupleField, Tuple


@addTupleType
class ImportDocumentTuple(Tuple):
    """Import Document Tuple

    This tuple is the publicly exposed Document

    """

    __tupleType__ = docDbTuplePrefix + "ImportDocumentTuple"

    #:  The unique key of this document
    key: str = TupleField()

    #:  The model set of this document
    modelSetKey: str = TupleField()

    #:  The document type
    documentTypeKey: str = TupleField()

    #:  The document data
    document: Dict = TupleField()

    #:  The hash of this import group
    importGroupHash: str = TupleField()

    def __repr__(self):
        s = ""
        s += f"key={self.key}\n"
        s += f"modelSetKey={self.modelSetKey}\n"
        s += f"documentTypeKey={self.documentTypeKey}\n"
        s += f"importGroupHash={self.importGroupHash}\n"
        for key, value in sorted(self.document.items(), key=lambda i: i[0]):
            s += f"prop.{key}={value}\n"
        return s



import logging
from typing import Dict

from peek_abstract_chunked_index.private.client.handlers.ACICacheHandlerABC import (
    ACICacheHandlerABC,
)
from peek_abstract_chunked_index.private.tuples.ACIUpdateDateTupleABC import (
    ACIUpdateDateTupleABC,
)
from peek_core_docdb._private.PluginNames import docDbFilt
from peek_core_docdb._private.client.controller.DocumentCacheController import (
    clientDocumentUpdateFromServerFilt,
)
from peek_core_docdb._private.tuples.DocumentUpdateDateTuple import (
    DocumentUpdateDateTuple,
)

logger = logging.getLogger(__name__)

clientDocumentWatchUpdateFromDeviceFilt = {
    "key": "clientDocumentWatchUpdateFromDevice"
}
clientDocumentWatchUpdateFromDeviceFilt.update(docDbFilt)


# ModelSet HANDLER
class DocumentCacheHandler(ACICacheHandlerABC):
    _UpdateDateTuple: ACIUpdateDateTupleABC = DocumentUpdateDateTuple
    _updateFromDeviceFilt: Dict = clientDocumentWatchUpdateFromDeviceFilt
    _updateFromLogicFilt: Dict = clientDocumentUpdateFromServerFilt
    _logger: logging.Logger = logger



from sqlalchemy import Column
from sqlalchemy import Integer, String
from vortex.Tuple import addTupleType, Tuple

from peek_core_docdb._private.PluginNames import docDbTuplePrefix
from .DeclarativeBase import DeclarativeBase


@addTupleType
class DocDbModelSet(Tuple, DeclarativeBase):
    __tablename__ = "DocDbModelSet"
    __tupleType__ = docDbTuplePrefix + __tablename__

    id = Column(Integer, primary_key=True, autoincrement=True)
    key = Column(String, nullable=False, unique=True)
    name = Column(String, nullable=False, unique=True)

    comment = Column(String)
    propsJson = Column(String)



from typing import Dict

from peek_abstract_chunked_index.private.tuples.ACIUpdateDateTupleABC import (
    ACIUpdateDateTupleABC,
)
from peek_core_docdb._private.PluginNames import docDbTuplePrefix
from vortex.Tuple import addTupleType, TupleField, Tuple


@addTupleType
class DocumentUpdateDateTuple(Tuple, ACIUpdateDateTupleABC):
    """DocDb Object Update Date Tuple

    This tuple represents the state of the chunks in the cache.
    Each chunkKey has a lastUpdateDate as a string, this is used for offline caching
    all the chunks.
    """

    __tupleType__ = docDbTuplePrefix + "DocumentUpdateDateTuple"

    # Improve performance of the JSON serialisation
    __rawJonableFields__ = ("initialLoadComplete", "updateDateByChunkKey")

    initialLoadComplete: bool = TupleField()
    updateDateByChunkKey: Dict[str, str] = TupleField({})

    @property
    def ckiUpdateDateByChunkKey(self):
        return self.updateDateByChunkKey

    def ckiSetUpdateDateByChunkKey(self, value: Dict[str, str]) -> None:
        self.updateDateByChunkKey = value

    @property
    def ckiInitialLoadComplete(self) -> bool:
        return self.initialLoadComplete

    def ckiSetInitialLoadComplete(self, value: bool) -> None:
        self.initialLoadComplete = value



import hashlib
import json
import logging
from base64 import b64encode
from collections import defaultdict
from datetime import datetime
from typing import List, Dict

import pytz
from sqlalchemy import select
from txcelery.defer import DeferrableTask
from vortex.Payload import Payload

from peek_plugin_base.worker import CeleryDbConn
from peek_plugin_base.worker.CeleryApp import celeryApp
from peek_core_docdb._private.storage.DocDbCompilerQueue import (
    DocDbCompilerQueue,
)
from peek_core_docdb._private.storage.DocDbDocument import DocDbDocument
from peek_core_docdb._private.storage.DocDbEncodedChunk import DocDbEncodedChunk

logger = logging.getLogger(__name__)

""" DocDb Index Compiler

Compile the docDb indexes

1) Query for queue
2) Process queue
3) Delete from queue
"""


@DeferrableTask
@celeryApp.task(bind=True)
def compileDocumentChunk(self, payloadEncodedArgs: bytes) -> dict[str, str]:
    """Compile DocDb Index Task

    :param self: The reference to this celery task
    :param payloadEncodedArgs: An encoded payload containing the queue tuples.
    :returns: A list of grid keys that have been updated.
    """
    argData = Payload().fromEncodedPayload(payloadEncodedArgs).tuples
    queueItems = argData[0]
    queueItemIds: List[int] = argData[1]

    engine = CeleryDbConn.getDbEngine()
    conn = engine.connect()
    transaction = conn.begin()
    try:
        queueItemsByModelSetId = defaultdict(list)

        for queueItem in queueItems:
            queueItemsByModelSetId[queueItem.modelSetId].append(queueItem)

        lastUpdateByChunkKey = {}
        for modelSetId, modelSetQueueItems in queueItemsByModelSetId.items():
            lastUpdateByChunkKey.update(
                _compileDocumentChunk(
                    conn, transaction, modelSetId, modelSetQueueItems
                )
            )

        queueTable = DocDbCompilerQueue.__table__

        transaction = conn.begin()
        conn.execute(queueTable.delete(queueTable.c.id.in_(queueItemIds)))
        transaction.commit()

    except Exception as e:
        transaction.rollback()
        logger.debug("RETRYING task - %s", e)
        raise self.retry(exc=e, countdown=10)

    finally:
        conn.close()

    return lastUpdateByChunkKey


def _compileDocumentChunk(
    conn, transaction, modelSetId: int, queueItems: List[DocDbCompilerQueue]
) -> dict[str, str]:
    chunkKeys = list(set([i.chunkKey for i in queueItems]))

    compiledTable = DocDbEncodedChunk.__table__
    lastUpdate = datetime.now(pytz.utc).isoformat()

    startTime = datetime.now(pytz.utc)

    logger.debug(
        "Staring compile of %s queueItems in %s",
        len(queueItems),
        (datetime.now(pytz.utc) - startTime),
    )

    # Get Model Sets

    total = 0
    existingHashes = _loadExistingHashes(conn, chunkKeys)
    encKwPayloadByChunkKey = _buildIndex(chunkKeys)
    chunksToDelete = []

    lastUpdateByChunkKey = {}

    inserts = []
    for (
        chunkKey,
        docDbIndexChunkEncodedPayload,
    ) in encKwPayloadByChunkKey.items():
        m = hashlib.sha256()
        m.update(docDbIndexChunkEncodedPayload)
        encodedHash = b64encode(m.digest()).decode()

        # Compare the hash, AND delete the chunk key
        if chunkKey in existingHashes:
            # At this point we could decide to do an update instead,
            # but inserts are quicker
            if encodedHash == existingHashes.pop(chunkKey):
                continue

        lastUpdateByChunkKey[str(chunkKey)] = lastUpdate

        chunksToDelete.append(chunkKey)
        inserts.append(
            dict(
                modelSetId=modelSetId,
                chunkKey=chunkKey,
                encodedData=docDbIndexChunkEncodedPayload,
                encodedHash=encodedHash,
                lastUpdate=lastUpdate,
            )
        )

    # Add any chnuks that we need to delete that we don't have new data for, here
    chunksToDelete.extend(list(existingHashes))

    if chunksToDelete:
        # Delete the old chunks
        conn.execute(
            compiledTable.delete(compiledTable.c.chunkKey.in_(chunksToDelete))
        )

    if inserts:
        newIdGen = CeleryDbConn.prefetchDeclarativeIds(
            DocDbDocument, len(inserts)
        )
        for insert in inserts:
            insert["id"] = next(newIdGen)

    transaction.commit()
    transaction = conn.begin()

    if inserts:
        conn.execute(compiledTable.insert(), inserts)

    logger.debug(
        "Compiled %s Documents, %s missing, in %s",
        len(inserts),
        len(chunkKeys) - len(inserts),
        (datetime.now(pytz.utc) - startTime),
    )

    total += len(inserts)

    transaction.commit()
    logger.info(
        "Compiled and Committed %s EncodedDocumentChunks in %s",
        total,
        (datetime.now(pytz.utc) - startTime),
    )

    return lastUpdateByChunkKey


def _loadExistingHashes(conn, chunkKeys: List[str]) -> Dict[str, str]:
    compiledTable = DocDbEncodedChunk.__table__

    results = conn.execute(
        select(
            columns=[compiledTable.c.chunkKey, compiledTable.c.encodedHash],
            whereclause=compiledTable.c.chunkKey.in_(chunkKeys),
        )
    ).fetchall()

    return {result[0]: result[1] for result in results}


def _buildIndex(chunkKeys) -> Dict[str, bytes]:
    session = CeleryDbConn.getDbSession()

    try:
        indexQry = (
            session.query(
                DocDbDocument.chunkKey,
                DocDbDocument.key,
                DocDbDocument.documentJson,
            )
            .filter(DocDbDocument.chunkKey.in_(chunkKeys))
            .order_by(DocDbDocument.key)
            .yield_per(1000)
            .all()
        )

        # Create the ChunkKey -> {id -> packedJson, id -> packedJson, ....]
        packagedJsonByObjIdByChunkKey = defaultdict(dict)

        for item in indexQry:
            packagedJsonByObjIdByChunkKey[item.chunkKey][
                item.key
            ] = item.documentJson

        encPayloadByChunkKey = {}

        # Sort each bucket by the key
        for chunkKey, packedJsonByKey in packagedJsonByObjIdByChunkKey.items():
            tuples = json.dumps(packedJsonByKey, sort_keys=True)

            # Create the blob data for this index.
            # It will be docDbed by a binary sort
            encPayloadByChunkKey[chunkKey] = (
                Payload(tuples=tuples).toEncodedPayload().encode()
            )

        return encPayloadByChunkKey

    finally:
        session.close()



import logging
from typing import Any

from twisted.internet.defer import inlineCallbacks
from txhttputil.site.FileUnderlayResource import FileUnderlayResource
from vortex.handler.TupleActionProcessorProxy import TupleActionProcessorProxy
from vortex.handler.TupleDataObservableProxyHandler import (
    TupleDataObservableProxyHandler,
)
from vortex.handler.TupleDataObserverClient import TupleDataObserverClient

from peek_core_docdb._private.PluginNames import docDbActionProcessorName
from peek_core_docdb._private.PluginNames import docDbFilt
from peek_core_docdb._private.PluginNames import docDbObservableName
from peek_core_docdb._private.client.TupleDataObservable import (
    makeClientTupleDataObservableHandler,
)
from peek_core_docdb._private.client.controller.DocumentCacheController import (
    DocumentCacheController,
)
from peek_core_docdb._private.client.handlers.DocumentCacheHandler import (
    DocumentCacheHandler,
)
from peek_core_docdb._private.storage.DeclarativeBase import loadStorageTuples
from peek_core_docdb._private.tuples import loadPrivateTuples
from peek_core_docdb.tuples import loadPublicTuples
from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_plugin_base.client.PluginClientEntryHookABC import (
    PluginClientEntryHookABC,
)

logger = logging.getLogger(__name__)


class ClientEntryHook(PluginClientEntryHookABC):
    def __init__(self, *args, **kwargs):
        """ " Constructor"""
        # Call the base classes constructor
        PluginClientEntryHookABC.__init__(self, *args, **kwargs)

        #: Loaded Objects, This is a list of all objects created when we start
        self._loadedObjects = []

    @classmethod
    def setupStaticWebResources(cls, platformApi: Any):
        from peek_plugin_base.client.PeekClientPlatformHookABC import (
            PeekClientPlatformHookABC,
        )

        assert isinstance(
            platformApi, PeekClientPlatformHookABC
        ), "We received the wrong platform API"

        from peek_abstract_chunked_index.private.client.controller.ACICacheControllerABC import (
            ACICacheControllerABC,
        )

        relativeDir = ACICacheControllerABC.appDownloadPluginDirRelativeDir()
        dirPath = platformApi.fileStorageDirectory / relativeDir
        dirPath.mkdir(exist_ok=True, parents=True)

        resource = FileUnderlayResource()
        resource.addFileSystemRoot(dirPath)

        platformApi.addFieldResource(relativeDir.encode(), resource)
        platformApi.addOfficeResource(relativeDir.encode(), resource)

    def load(self) -> None:
        """Load

        This will be called when the plugin is loaded,
         just after the db is migrated.
        Place any custom initialiastion steps here.

        """

        loadStorageTuples()

        loadPrivateTuples()
        loadPublicTuples()

        logger.debug("Loaded")

    @inlineCallbacks
    def start(self):
        """Load

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialisation steps here.

        """

        # ----------------
        # Proxy actions back to the server, we don't process them at all
        self._loadedObjects.append(
            TupleActionProcessorProxy(
                tupleActionProcessorName=docDbActionProcessorName,
                proxyToVortexName=peekServerName,
                additionalFilt=docDbFilt,
            )
        )

        # ----------------
        # Provide the devices access to the servers observable
        tupleDataObservableProxyHandler = TupleDataObservableProxyHandler(
            observableName=docDbObservableName,
            proxyToVortexName=peekServerName,
            additionalFilt=docDbFilt,
            observerName="Proxy to devices",
        )
        self._loadedObjects.append(tupleDataObservableProxyHandler)

        # ----------------
        #: This is an observer for us (the client) to use to observe data
        # from the server
        serverTupleObserver = TupleDataObserverClient(
            observableName=docDbObservableName,
            destVortexName=peekServerName,
            additionalFilt=docDbFilt,
            observerName="Data for client",
        )
        self._loadedObjects.append(serverTupleObserver)

        # ----------------
        # Document Cache Controller

        documentCacheController = DocumentCacheController(
            self.platform.serviceId, self.platform.fileStorageDirectory
        )
        self._loadedObjects.append(documentCacheController)

        # ----------------
        # Document Cache Handler

        documentHandler = DocumentCacheHandler(
            cacheController=documentCacheController,
            clientId=self.platform.serviceId,
        )
        self._loadedObjects.append(documentHandler)

        # ----------------
        # Set the caches reference to the handler
        documentCacheController.setCacheHandler(documentHandler)

        # ----------------
        # Create the Tuple Observer
        makeClientTupleDataObservableHandler(
            tupleDataObservableProxyHandler, documentCacheController
        )

        # ----------------
        # Start the compiler controllers
        yield documentHandler.start()
        yield documentCacheController.start()

        logger.debug("Started")

    def stop(self):
        """Stop

        This method is called by the platform to tell the peek app to shutdown and stop
        everything it's doing
        """
        # Shutdown and dereference all objects we constructed when we started
        while self._loadedObjects:
            self._loadedObjects.pop().shutdown()

        logger.debug("Stopped")

    def unload(self):
        """Unload

        This method is called after stop is called, to unload any last resources
        before the PLUGIN is unlinked from the platform

        """
        logger.debug("Unloaded")



import logging

from peek_core_docdb._private.PluginNames import docDbFilt
from peek_core_docdb._private.storage.DocDbPropertyTuple import DocDbPropertyTuple
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler
from vortex.sqla_orm.OrmCrudHandler import OrmCrudHandler, OrmCrudHandlerExtension

logger = logging.getLogger(__name__)

# This dict matches the definition in the Admin angular app.
filtKey = {"key": "admin.Edit.DocDbPropertyTuple"}
filtKey.update(docDbFilt)


# This is the CRUD hander‰
class __CrudHandler(OrmCrudHandler):
    pass


class __ExtUpdateObservable(OrmCrudHandlerExtension):
    """Update Observable ORM Crud Extension

    This extension is called after events that will alter data,
    it then notifies the observer.

    """

    def __init__(self, tupleDataObserver: TupleDataObservableHandler):
        self._tupleDataObserver = tupleDataObserver

    def _tellObserver(self, tuple_, tuples, session, payloadFilt):
        self._tupleDataObserver.notifyOfTupleUpdate(
            TupleSelector(DocDbPropertyTuple.tupleName(), {})
        )
        return True

    afterUpdateCommit = _tellObserver
    afterDeleteCommit = _tellObserver


# This method creates an instance of the handler class.
def makeDocumentPropertyHandler(tupleObservable, dbSessionCreator):
    handler = __CrudHandler(
        dbSessionCreator, DocDbPropertyTuple, filtKey, retreiveAll=True
    )

    logger.debug("Started")
    handler.addExtension(DocDbPropertyTuple, __ExtUpdateObservable(tupleObservable))
    return handler



import logging
from typing import Dict

from peek_abstract_chunked_index.private.server.client_handlers.ACIChunkUpdateHandlerABC import (
    ACIChunkUpdateHandlerABC,
)
from peek_abstract_chunked_index.private.tuples.ACIEncodedChunkTupleABC import (
    ACIEncodedChunkTupleABC,
)
from peek_core_docdb._private.client.controller.DocumentCacheController import (
    clientDocumentUpdateFromServerFilt,
)
from peek_core_docdb._private.storage.DocDbEncodedChunk import DocDbEncodedChunk

logger = logging.getLogger(__name__)


class ClientChunkUpdateHandler(ACIChunkUpdateHandlerABC):
    _ChunkedTuple: ACIEncodedChunkTupleABC = DocDbEncodedChunk
    _updateFromLogicFilt: Dict = clientDocumentUpdateFromServerFilt
    _logger: logging.Logger = logger



"""Added showInHeader

Peek Plugin Database Migration Script

Revision ID: 6c94c040e18c
Revises: 931b88db3117
Create Date: 2019-07-30 13:55:15.386606

"""

# revision identifiers, used by Alembic.
revision = "6c94c040e18c"
down_revision = "931b88db3117"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "DocDbProperty",
        sa.Column("showInHeader", sa.Boolean(), nullable=False, server_default="0"),
        schema="core_docdb",
    )

    op.execute(""" UPDATE "core_docdb"."DocDbProperty" SET "showInHeader" = false """)

    op.alter_column(
        "DocDbProperty",
        "showInHeader",
        type_=sa.Boolean(),
        nullable=False,
        schema="core_docdb",
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("DocDbProperty", "showInHeader", schema="core_docdb")
    # ### end Alembic commands ###



from sqlalchemy import Column, Index, ForeignKey, BigInteger
from sqlalchemy import Integer, String
from sqlalchemy.orm import relationship

from peek_core_docdb._private.PluginNames import docDbTuplePrefix
from peek_core_docdb._private.storage.DeclarativeBase import DeclarativeBase
from peek_core_docdb._private.storage.DocDbDocumentTypeTuple import (
    DocDbDocumentTypeTuple,
)
from peek_core_docdb._private.storage.DocDbModelSet import DocDbModelSet
from vortex.Tuple import Tuple, addTupleType


@addTupleType
class DocDbDocument(Tuple, DeclarativeBase):
    __tupleType__ = docDbTuplePrefix + "DocDbDocumentTable"
    __tablename__ = "DocDbDocument"

    #:  The unique ID of this document (database generated)
    id = Column(BigInteger, primary_key=True, autoincrement=True)

    #:  The model set for this document
    modelSetId = Column(
        Integer, ForeignKey("DocDbModelSet.id", ondelete="CASCADE"), nullable=False
    )
    modelSet = relationship(DocDbModelSet)

    #:  The model set for this document
    documentTypeId = Column(
        Integer, ForeignKey("DocDbDocumentType.id", ondelete="CASCADE"), nullable=False
    )
    documentType = relationship(DocDbDocumentTypeTuple)

    importGroupHash = Column(String, nullable=False)

    #:  The unique key of this document
    key = Column(String, nullable=False)

    #:  The chunk that this document fits into
    chunkKey = Column(String, nullable=False)

    #:  The document data
    documentJson = Column(String, nullable=False)

    __table_args__ = (
        Index("idx_Document_key", modelSetId, key, unique=True),
        Index("idx_Document_documentType", documentTypeId, unique=False),
        Index("idx_Document_gridKey", chunkKey, unique=False),
        Index("idx_Document_importGroupHash", importGroupHash, unique=False),
    )



import logging

from twisted.internet.defer import inlineCallbacks

from peek_core_docdb._private.worker.tasks.ImportTask import createOrUpdateDocuments

logger = logging.getLogger(__name__)


class ImportController:
    def __init__(self):
        pass

    def shutdown(self):
        pass

    @inlineCallbacks
    def createOrUpdateDocuments(self, documentsEncodedPayload: bytes):
        yield createOrUpdateDocuments.delay(documentsEncodedPayload)
